# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from typing import Optional

from frm.enums import DayCountBasis
from frm.utils import clean_tenor, tenor_to_date_offset, year_frac, get_busdaycal, workday


def cash_rates_quote_helper(
        df: pd.DataFrame,
        curve_date: pd.Timestamp,
        day_count_basis: DayCountBasis,
        settlement_delay: Optional[int]=None,
        settlement_date: Optional[pd.Timestamp]=None,
        cal: np.busdaycalendar=np.busdaycalendar):

    assert 'rate' in df.columns
    assert 'tenor' in df.columns or ('effective_date' in df.columns and 'maturity_date' in df.columns)

    df['tenor'] = df['tenor'].apply(clean_tenor)

    if settlement_date is None:
        assert settlement_delay is not None
        settlement_date = workday(dates=curve_date, offset=settlement_delay, cal=cal)

    if 'effective_date' not in df.columns:
        df['has_settlement_delay'] = df['tenor'] != 'on'
        df.loc[~df['has_settlement_delay'],'effective_date'] = curve_date
        df.loc[df['has_settlement_delay'],'effective_date'] = settlement_date

    if 'maturity_date' not in df.columns:
        date_offset = df['tenor'].apply(tenor_to_date_offset).values
        df['maturity_date'] = workday(dates=(df['effective_date'] + date_offset), offset=0, cal=cal)

    df['period_years'] = year_frac(df['effective_date'], df['maturity_date'], day_count_basis)

    return df

def bootstrap_cash_rates(
        df: pd.DataFrame,
        curve_date: pd.Timestamp,
        day_count_basis_years: DayCountBasis=DayCountBasis.ACT_ACT):

    df['effective_date_years'] = year_frac(curve_date, df['effective_date'], day_count_basis_years)
    df['maturity_date_years'] = year_frac(curve_date, df['maturity_date'], day_count_basis_years)
    df['effective_date_discount_factor'] = np.nan
    df['maturity_date_discount_factor'] = np.nan

    df.sort_values(by='maturity_date', inplace=True)
    for i, row in df.iterrows():
        if not row['has_settlement_delay']:
            df.at[i, 'effective_date_discount_factor'] = 1.0
            df.at[i, 'maturity_date_discount_factor'] = 1.0 / (1 + row['rate'] * row['period_years'])
        else:
            mask = df['maturity_date'] <= row['effective_date']

            x = row['effective_date_years']
            X = df.loc[mask, 'maturity_date_years'].values
            Y = np.log(df.loc[mask, 'maturity_date_discount_factor'].values)
            interp_func = interp1d(X, Y, kind='linear')
            df.at[i, 'effective_date_discount_factor'] = np.exp(interp_func(x))

            df.at[i, 'maturity_date_discount_factor'] = df.at[i, 'effective_date_discount_factor'] \
                                                                / (1 + row['rate'] * row['period_years'])

    #df = df.drop(columns=['effective_date_years', 'maturity_date_years'], errors='ignore')

    return df



def bootstrap_futures(
        df: pd.DataFrame,
        curve_date: pd.Timestamp,
        cash_rates_df: pd.DataFrame,
        day_count_basis: DayCountBasis,
        day_count_basis_years: DayCountBasis=DayCountBasis.ACT_ACT):

    if 'futures_rate' not in df.columns:
        df['futures_rate'] = (100 - df['price']) / 100
    if 'convexity_bps' not in df.columns:
        df.insert(df.columns.get_loc('price') + 1, 'convexity_bps', 0)
    if 'forward_rate' not in df.columns:
        df['forward_rate'] = df['futures_rate'] - df['convexity_bps'] / 10000

    df['period_years'] = year_frac(df['effective_date'], df['maturity_date'], day_count_basis)
    df['effective_date_years'] = year_frac(curve_date, df['effective_date'], day_count_basis_years)
    df['maturity_date_years'] = year_frac(curve_date, df['maturity_date'], day_count_basis_years)

    X = cash_rates_df['maturity_date_years'].values
    Y = np.log(cash_rates_df['maturity_date_discount_factor'].values)

    df.sort_values(by='effective_date', inplace=True)
    df.reset_index(inplace=True, drop=True)
    for i, row in df.iterrows():
        x = row['effective_date_years']
        interp_func = interp1d(X, Y, kind='linear', fill_value='extrapolate')
        df.at[i, 'effective_date_discount_factor'] = np.exp(interp_func(x))

        df.at[i, 'maturity_date_discount_factor'] = df.at[i, 'effective_date_discount_factor'] \
                                                    / (1 + row['forward_rate'] * row['period_years'])

        X = np.append(X, row['maturity_date_years'])
        Y = np.append(Y, np.log(df.at[i, 'maturity_date_discount_factor']))

    return df


def futures_helper_asx_90day(
        df: pd.DataFrame
    ):
    # Helper function to setup contract details for ASX 90-day bank bill futures.
    # https://www.asx.com.au/content/dam/asx/participants/derivatives-market/90-Day-bank-bill-futures-factsheet.pdf

    # From the contract code, index the effective period of the BBSW 3M fixing of the contract.

    if 'months' not in df.columns or 'years' not in df.columns:

        if 'ticker' in df.columns:

            df['ticker'] = df['ticker'].str.upper().str.strip()
            contract_code = df['ticker'].str[:2]
            assert (contract_code == 'IR').all() # 'IR' for 90-day bank bill futures

            futures_month_code = df['ticker'].str[2]
            assert futures_month_code.isin(['H', 'M', 'U', 'Z']).all() # 'H', 'M', 'U', 'Z' for Mar, Jun, Sep, Dec
            df['months'] = futures_month_code.map({'H': 3, 'M': 6, 'U': 9, 'Z': 12})

            assert df['ticker'].str.len().isin([1, 2, 4]).all()
            futures_year_code = df['ticker'].str[3:].astype(int)  # 1-4-digit year code

            def resolve_year(year_code):
                year_code_str = str(year_code)
                if len(year_code_str) == 1:
                    # Map single-digit year codes to 2020-2029
                    # TODO: does not support cases where the futures schedule spans multiple decades. i.e from 2028-2031.
                    return 2020 + int(year_code_str)
                elif len(year_code_str) == 2:
                    # Map two-digit year codes to current century
                    return 2000 + int(year_code_str)
                elif len(year_code_str) == 4:
                    # Use four-digit year codes as is
                    return int(year_code_str)
                else:
                    raise ValueError(f'Unexpected year code length: {year_code_str}')

            df['years'] = futures_year_code.apply(resolve_year)

        elif 'month_year' in df.columns:
            # SEP23, DEC23, MAR24, JUN24...
            # SEP2024, DEC2024, MAR2025, JUN2025...
            pass



    expiry_dates, settlement_dates = [], []
    for years in range(df['years'].min(), df['years'].max()+1):
        for months in [3, 6, 9, 12]:
            if ((df['years'] == years) & (df['months'] == months)).any():
                # Settlement Day = 2nd Friday of each delivery month (Mar, Jun, Sep, Dec).
                # Start from the 1st of the month and get all Fridays in the month
                month_start = pd.Timestamp(year=years, month=months, day=1)
                fridays = [d for d in pd.date_range(month_start, month_start + pd.offsets.MonthEnd(), freq='W-FRI')]
                settlement_date = fridays[1] # Get the second Friday
                settlement_dates.append(settlement_date)

                # Expiry date is the business day before settlement date
                expiry_dates.append(settlement_date - pd.offsets.BDay(1))

    df['expiry_date'] = expiry_dates
    df['settlement_date'] = settlement_dates

    # Calculate the effective period of the BBSW 3M fixing of the contract.
    df['effective_date'] = df['settlement_date']
    cal = get_busdaycal('AU-SYDNEY')
    df['maturity_date'] = workday(df['settlement_date'] + pd.DateOffset(months=3), 0, cal=cal)

    if 'price' in df.columns:
        # Per the ASX contract definition: Settlement Value = 100 - [BBSW 3M Fixing]
        # Hence: BBSW 3M Fixing implied by the price is 100 - price.
        df['futures_rate'] = (100 - df['price']) * 0.01

    return df





# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))

import pandas as pd

def clean_enum_value(value, transform_fn=None):
    if isinstance(value, str):
        value = value.lower().strip().replace(' ','_')
        if transform_fn:
            value = transform_fn(value)
        if value.isdigit():
            value = int(value)
    elif pd.isna(value) or value is None:
        value = None
    return value
        
    
def is_valid_enum_value(enum_class, value):
    value = clean_enum_value(value)
    return value in {enum_member.value for enum_member in enum_class}


def get_enum_member(enum_class, value, transform_fn=None):
    cleaned_value= clean_enum_value(value, transform_fn)
    if cleaned_value is None:
        return enum_class.default()
    for enum_member in enum_class:
        if enum_member.value == cleaned_value:
            return enum_member

    # List all valid codes in case of an error
    valid_values = [enum_member.value for enum_member in enum_class]
    raise ValueError(f"Invalid value: {value}. Valid codes are: {valid_values}")
        
        





# -*- coding: utf-8 -*-
from functools import reduce
import os
import holidays
import numpy as np
import pandas as pd
import dill
import openpyxl

# Set the base directory to the 'frm/utils' directory relative to the current module location
utils_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '.'))
file_path_locale_holiday = os.path.join(utils_dir, 'LOCALE_HOLIDAY.pkl')
file_path_ccy_holiday = os.path.join(utils_dir, 'CCY_HOLIDAY.pkl')

with open(file_path_locale_holiday, 'rb') as f:
    LOCALE_HOLIDAY = dill.load(f)

with open(file_path_ccy_holiday, 'rb') as f:
    CCY_HOLIDAY = dill.load(f)


def get_holidays_object(key):
    if key in CCY_HOLIDAY.keys():
        return CCY_HOLIDAY[key]
    elif key in LOCALE_HOLIDAY.keys():
        return LOCALE_HOLIDAY[key]
    else:
        raise ValueError(f"Holidays not setup for {key}")


def convert_weekend_to_weekmask(weekend_set):
    "For converting the holidays.weekend attribute to a valid weekmask for numpy.busdaycalendar"
    # Initialize a list of ones representing Monday to Sunday (working days)
    weekmask = [1, 1, 1, 1, 1, 1, 1]
    
    # Loop through the weekend days and set corresponding indexes in the weekmask to 0
    for weekend_day in weekend_set:
        weekmask[weekend_day] = 0
    
    # Return the weekmask as a string (optional, can also return as a list if needed)
    return ''.join(map(str, weekmask))


def get_busdaycal(keys) -> np.busdaycalendar:
    """
    Create a calendar which has the holidays and business days of the currencies/locales.
    """

    if keys is None:
        return np.busdaycalendar()

    if type(keys) is str:
        keys = [keys.upper()]
    elif type(keys) is list:
        keys = list(set(keys))
        keys = [key.upper() for key in keys]

    holiday_objects = [get_holidays_object(key) for key in keys]
    combined_weekend_union = reduce(lambda x, y: x | y.weekend, holiday_objects, set())
    weekmasks = convert_weekend_to_weekmask(combined_weekend_union)

    # Flattan the list of lists
    holidays = [h for holiday_list in holiday_objects for h in holiday_list]

    return np.busdaycalendar(weekmask=weekmasks, holidays=holidays)  



# Pickle variables 
if __name__ == "__main__":
    
    years = range(1990,2100)

    LOCALE_HOLIDAY = dict(sorted({
        'AE-DUBAI': holidays.AE(categories=['public'], years=years),
        'AR-BUENOS_AIRES': holidays.AR(categories=['public'], years=years),
        'AU-SYDNEY': holidays.AU(subdiv='NSW', categories=['public', 'bank'], years=years),
        'AU-MELBOURNE': holidays.AU(subdiv='VIC', categories=['public', 'bank'], years=years),
        'BR-SAO_PAULO': holidays.BR(subdiv='SP', categories=['public'], years=years),
        'CA-TORONTO': holidays.CA(subdiv='ON', categories=['public'], years=years),
        'CH-ZURICH': holidays.CH(subdiv='ZH', categories=['public'], years=years),
        'CL-SANTIAGO': holidays.CL(subdiv='RM', categories=['public', 'bank'], years=years),
        'CN-SHANGHAI': holidays.CN(categories=['public'], years=years),
        'CO-BOGOTA': holidays.CO(categories=['public'], years=years),
        'CZ-PRAGUE': holidays.CZ(categories=['public'], years=years),
        'DK-COPENHAGEN': holidays.DK(categories=['public'], years=years),
        'DE-FRANKFURT': holidays.DE(subdiv='HH', categories=['public'], years=years),
        'ES-MADRID': holidays.ES(subdiv='MD', categories=['public'], years=years),
        'FR-PARIS': holidays.FR(categories=['public'], years=years),
        'FI-HELSINKI': holidays.FI(categories=['public'], years=years),
        'HK-HONG_KONG': holidays.HK(categories=['public'], years=years),
        'HU-BUDAPEST': holidays.HU(categories=['public'], years=years),
        'ID-JAKARTA': holidays.ID(categories=['public'], years=years),
        'IN-MUMBAI': holidays.IN(subdiv='MH', categories=['public'], years=years),
        'IL-TEL_AVIV': holidays.IL(categories=['public'], years=years),
        'IS-REYKJAVIK': holidays.IS(categories=['public'], years=years),
        'JP-TOYKO': holidays.JP(categories=['public'], years=years),
        'MY-KUALA_LUMPUR': holidays.MY(subdiv='KUL', categories=['public'], years=years),
        'MX-MEXICO_CITY': holidays.MX(categories=['public'], years=years),
        'NO-OSLO': holidays.NO(categories=['public'], years=years),
        'NZ-AUCKLAND': holidays.NZ(subdiv='AUK', categories=['public'], years=years),
        'PL-WARSAW': holidays.PL(categories=['public'], years=years),
        'PH-MANILA': holidays.PH(categories=['public'], years=years),
        'RU-MOSCOW': holidays.RU(categories=['public'], years=years),
        'SA-RIYADH': holidays.SA(categories=['public'], years=years),
        'SE-STOCKHOLM': holidays.SE(categories=['public'], years=years),
        'SG-SINGAPORE': holidays.SG(categories=['public'], years=years),
        'TH-BANKCOK': holidays.TH(categories=['public', 'bank'], years=years),
        'TR-ISTANBUL': holidays.TR(categories=['public'], years=years),
        'TW-TAPEI': holidays.TW(categories=['public'], years=years),
        'UK-LONDON': holidays.UK(subdiv='ENG', categories=['public'], years=years),
        'US-NEW_YORK': holidays.US(subdiv='NY', categories=['public'], years=years),
        'VN-HO_CHI_MINH_CITY': holidays.VN(categories=['public'], years=years),
        'ZA-JOHANNESBURG': holidays.ZA(categories=['public'], years=years),
        'European_Central_Bank': holidays.ECB(categories=['public'], years=years),
        'ICE_Futures_Europe': holidays.IFEU(categories=['public'], years=years), # Requires holidays >=0.49
        'New_York_Stock_Exchange': holidays.XNYS(categories=['public'], years=years),
    }.items()))

    CCY_HOLIDAY = dict(sorted({
        'AED': holidays.AE(categories=['public'], years=years),
        'ARS': holidays.AR(categories=['public'], years=years),
        'AUD': holidays.AU(subdiv='NSW', categories=['public', 'bank'], years=years),
        'BRL': holidays.BR(subdiv='SP', categories=['public'], years=years),
        'CAD': holidays.CA(subdiv='ON', categories=['public'], years=years),
        'CHF': holidays.CH(subdiv='ZH', categories=['public'], years=years),
        'CLP': holidays.CL(subdiv='RM', categories=['public', 'bank'], years=years),
        'CNY': holidays.CN(categories=['public'], years=years),
        'COP': holidays.CO(categories=['public'], years=years),
        'CZK': holidays.CZ(categories=['public'], years=years),
        'DKK': holidays.DK(categories=['public'], years=years),
        'HKD': holidays.HK(categories=['public'], years=years),
        'HUF': holidays.HU(categories=['public'], years=years),
        'IDR': holidays.ID(categories=['public'], years=years),
        'INR': holidays.IN(subdiv='MH', categories=['public'], years=years),
        'ILS': holidays.IL(categories=['public'], years=years),
        'ISK': holidays.IS(categories=['public'], years=years),
        'JPY': holidays.JP(categories=['public'], years=years),
        'MYR': holidays.MY(subdiv='KUL', categories=['public'], years=years),
        'MXN': holidays.MX(categories=['public'], years=years),
        'NOK': holidays.NO(categories=['public'], years=years),
        'NZD': holidays.NZ(subdiv='AUK', categories=['public'], years=years),
        'PLN': holidays.PL(categories=['public'], years=years),
        'PHP': holidays.PH(categories=['public'], years=years),
        'RUB': holidays.RU(categories=['public'], years=years),
        'SAR': holidays.SA(categories=['public'], years=years),
        'SEK': holidays.SE(categories=['public'], years=years),
        'SGD': holidays.SG(categories=['public'], years=years),
        'THB': holidays.TH(categories=['public', 'bank'], years=years),
        'TRY': holidays.TR(categories=['public'], years=years),
        'TWD': holidays.TW(categories=['public'], years=years),
        'GBP': holidays.UK(subdiv='ENG', categories=['public'], years=years),
        'USD': holidays.US(subdiv='NY', categories=['public'], years=years),
        'VND': holidays.VN(categories=['public'], years=years),
        'ZAR': holidays.ZA(categories=['public'], years=years),
        'EUR': holidays.ECB(categories=['public'], years=years),
    }.items()))

    # Pickle the variable
    with open(file_path_locale_holiday, 'wb') as f:
        dill.dump(LOCALE_HOLIDAY, f)

    # Pickle the variable
    with open(file_path_ccy_holiday, 'wb') as f:
        dill.dump(CCY_HOLIDAY, f)
        

if __name__ == "__main__":

    # Create code for static holidays definition
    df = pd.read_excel(utils_dir + "\\calendar_map.xlsx", usecols=range(6))
    holiday_dict = {}
    holiday_dict_str = "LOCALE_HOLIDAY = dict(sorted({\n"
    ccy_holiday_dict = {}
    ccy_holiday_dict_str = "CCY_HOLIDAY = dict(sorted({\n"

    for i, row in df.iterrows():
        eval_str = None
        ccy = row["ISOCurrencyCode"]
        ui_string = row["UIString"]
        code = row["Code"]
        subdivision = row["Subdivision"]

        try:
            categories = [holidays.PUBLIC, holidays.BANK]
            if pd.isna(row["Subdivision"]):
                eval_str = f"holidays.{code}(categories={categories}, years=years)"

            else:
                eval_str = (
                    f"holidays.{code}(subdiv='{subdivision}', categories={categories}, years=years)"
                )
            holiday_obj = eval(eval_str)

        except ValueError:
            categories = [holidays.PUBLIC]
            if pd.isna(row["Subdivision"]):
                eval_str = f"holidays.{code}(categories={categories}, years=years)"
            else:
                eval_str = (
                    f"holidays.{code}(subdiv='{subdivision}', categories={categories}, years=years)"
                )
            holiday_obj = eval(eval_str)

        holiday_dict[ui_string] = holiday_obj
        holiday_dict_str += f"    '{ui_string}': {eval_str},\n"

        if row["CurrencyPrimaryCalendar"]:
            ccy_holiday_dict[ccy] = holiday_obj
            ccy_holiday_dict_str += f"    '{ccy}': {eval_str},\n"

    holiday_dict_str += "}.items()))"
    ccy_holiday_dict_str += "}.items()))"

    print(holiday_dict_str)
    print("\n")
    print(ccy_holiday_dict_str)




    # Initialize Excel writer
    # Prepare the output file path
    output_path = os.path.join(utils_dir, 'currency_holidays.xlsx')

    # Create a new workbook and select the active worksheet
    wb = openpyxl.Workbook()
    ws = wb.active
    ws.title = 'holidays'

    # Define column mappings for each currency
    col_start = 1  # Initial column start in openpyxl (1-based index)

    for ccy in ['aud', 'eur', 'gbp', 'jpy', 'nzd', 'usd']:
        # Prepare data for each currency
        dates = list(CCY_HOLIDAY[ccy.upper()].keys())
        holiday_name = list(CCY_HOLIDAY[ccy.upper()].values())

        # Create a DataFrame with currency-specific columns
        holiday_df = pd.DataFrame({
            f'{ccy.upper()}_date': dates,
            f'{ccy.upper()}_holiday_name': holiday_name
        })

        # Sort the DataFrame by date
        holiday_df.sort_values(by=f'{ccy.upper()}_date', inplace=True)
        holiday_df.reset_index(drop=True, inplace=True)

        # Write headers
        ws.cell(row=1, column=col_start, value=f'{ccy.upper()}_date')
        ws.cell(row=1, column=col_start + 1, value=f'{ccy.upper()}_holiday_name')

        # Write sorted data
        for i, row in holiday_df.iterrows():
            ws.cell(row=i + 2, column=col_start, value=row[f'{ccy.upper()}_date'])
            ws.cell(row=i + 2, column=col_start + 1, value=row[f'{ccy.upper()}_holiday_name'])

        # Move to the next pair of columns for the next currency
        col_start += 2

    # Save the workbook
    wb.save(output_path)

    # Define the output path
    #output_path = os.path.join(utils_dir, 'currency_holidays.xlsx')

    # Write to an Excel file, starting from column A
    #combined_df.to_excel(output_path, index=False, sheet_name='Currency Holidays')


# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from dataclasses import dataclass
import numpy as np
import os
import pandas as pd
from typing import Optional
import warnings

import frm.utils
from frm.term_structures.swap_curve import TermSwapCurve, RFRSwapCurve
from frm.utils import CouponSchedule, year_frac, MarketDataNotAvailableError
from frm.enums import CompoundingFreq, TermRate, RFRFixingCalcMethod, DayCountBasis, PayRcv
from frm.term_structures.zero_curve import ZeroCurve
from frm.term_structures.zero_curve_helpers import discount_factor_from_zero_rate
from scipy.optimize import root_scalar


if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))

# I am writing a (python) class to support the valuation of generic swap legs and bonds.

# THe key methods for a 'leg' are
# (i) coupon date schedule → schedule with forward coupon cashflow amount (will be different for each leg type)
# (ii) discount cashflows & get PV (this will be identical for all leg types)
# (iii) solve parameter (coupon rate, spread) in order to match leg to given price/value.
# (iv) clean mv, dirty mv, accrued interest, cashflow buckets, current, non-current, etc.
# (v) DV01, +/- 100 bps shift to par rates

# Leg types:
# Fixed, Zerocoupon
# FloatTerm, FloatRFR, InflationZC, InflationYoY
# Each leg type will have a different method for calculating the forward coupon cashflow amount.



@dataclass
class Leg(ABC):
    "Generic class to cover pricing of all types of swap legs and bonds"

    schedule: CouponSchedule
    discount_curve: Optional[ZeroCurve]=None
    pay_rcv: PayRcv=PayRcv.PAY
    day_count_basis: DayCountBasis=DayCountBasis.ACT_365
    notional_ccy: str=None
    settlement_ccy: str=None

    def __post_init__(self):
        self.schedule.df.loc[:,['notional', 'notional_payment']] = self.schedule.df[['notional', 'notional_payment']] * self.pay_rcv.multiplier
        self.schedule.add_period_daycount(day_count_basis=self.day_count_basis)
        self.schedule.add_period_yearfrac(day_count_basis=self.day_count_basis)

    @abstractmethod
    def calc_coupon_payment(self, coupon_override=None):
        pass

    @abstractmethod
    def _calc_accrued_interest(self):
        pass

    @abstractmethod
    def _get_solved_field_name(self):
        pass

    @abstractmethod
    def _get_legtype_specific_schedule_cols(self):
        pass

    def _add_contractual_coupon_param(self, coupon_param):
        """Validate and apply the coupon_contractual_component to the notional schedule."""
        coupon_param = np.atleast_1d(coupon_param)
        index, valid_shapes = self.schedule.determine_valid_shapes_for_coupon_param()
        coupon_param_name = self._get_solved_field_name()

        if coupon_param.shape not in valid_shapes:
            raise ValueError(
                f"Invalid {coupon_param_name} shape: {coupon_param.shape}. "
                f"Expected one of: {valid_shapes}"
            )

        index_coupon_payment_date = self.schedule.df.columns.get_loc('coupon_payment_date')
        self.schedule.df.insert(index_coupon_payment_date, coupon_param_name, np.nan)
        if coupon_param.shape == (1,):
            self.schedule.df.loc[index, coupon_param_name] = coupon_param[0]
        else:
            self.schedule.df.loc[index, coupon_param_name] = coupon_param


    def _get_contractual_schedule_cols(self):

        abstract_columns = ['period_start', 'period_end', 'notional_payment_date', 'coupon_payment_date']
        legtype_specific_columns = self._get_legtype_specific_schedule_cols()

        cols = []
        for col in self.schedule.df.columns:
            if col in abstract_columns or col in legtype_specific_columns:
                cols.append(col)
        return cols

    def target_solve(self, target_value):
        solved_value = self._solver_helper(solved_fieldname=self._get_solved_field_name(), target_value=target_value)
        return solved_value


    def par_solve(self):
        if 'notional_payment_date' not in self.schedule.df.columns:
            par_value = 0
        else:
            if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
                mask = self.schedule.df['notional_payment_date'] >= frm.utils.VALUE_DATE
            else:
                mask = self.schedule.df['notional_payment_date'] > frm.utils.VALUE_DATE
            # Par value is the outstanding notional payments.
            par_value = self.schedule.df['notional_payment'][mask].sum()

        solved_value = self._solver_helper(solved_fieldname=self._get_solved_field_name(), target_value=par_value)
        return solved_value

    def _solver_helper(self, solved_fieldname, target_value):
        field_original_value = self.schedule.df[solved_fieldname].values.copy()

        def objective_function(overide):
            return self.calc_dirty_pv(coupon_override=overide) - target_value

        solved_value = np.nan
        try:
            bracket = [-0.1, 3] # -10% to 300%
            solution = root_scalar(objective_function, bracket=bracket, method='brentq')
            if solution.converged:
                solved_value = solution.root
        finally:
            # Ensure reset, even if the solver fails
            self.schedule.df[solved_fieldname] = field_original_value

            return solved_value

    def calc_accrued_interest(self):
        # Leg specific accrued interest calculation
        accrued_interest = self._calc_accrued_interest()

        # Calculate the unsettled cashflows
        if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
            mask = (self.schedule.df['coupon_payment_date'] >= frm.utils.VALUE_DATE) & (self.schedule.df['period_end'] <= frm.utils.VALUE_DATE)
        else:
            mask = (self.schedule.df['coupon_payment_date'] > frm.utils.VALUE_DATE) & (self.schedule.df['period_end'] <= frm.utils.VALUE_DATE)
        if mask.sum() > 1:
            warnings.warn('Multiple periods found for unsettled cashflow calculation.')
        unsettled_cashflows = self.schedule.df['coupon_payment'].sum()

        if frm.utils.LIMIT_ACCRUED_INTEREST_TO_UNSETTLED_CASHFLOW:
            if unsettled_cashflows > 0:
                return unsettled_cashflows
            else:
                return accrued_interest
        else:
            return accrued_interest + unsettled_cashflows

    def _discount_payments(self, prefix=''):
        if self.discount_curve is None:
            raise MarketDataNotAvailableError('Discount curve not available.')

        payment_date_col = prefix + 'payment_date'
        discount_factor_col = prefix + 'discount_factor'

        if payment_date_col in self.schedule.df.columns:
            dates = self.schedule.df[payment_date_col]
            self.schedule.df[discount_factor_col] = self.discount_curve.get_discount_factors(self.schedule.df[payment_date_col])
            self.schedule.df.loc[pd.isna(dates),discount_factor_col] = 0.0

        if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
            mask = self.schedule.df[payment_date_col] < frm.utils.VALUE_DATE
        else:
            mask = self.schedule.df[payment_date_col] <= frm.utils.VALUE_DATE
        self.schedule.df.loc[mask, discount_factor_col] = 0.0

    def calc_dirty_pv(self, coupon_override=None):
        self.calc_coupon_payment(coupon_override=coupon_override)
        self.schedule.df['total_payment'] = self.schedule.df['notional_payment'] + self.schedule.df['coupon_payment']
        self._discount_payments(prefix='notional_')
        self._discount_payments(prefix='coupon_')
        self.schedule.df['notional_payment_pv'] = (self.schedule.df['notional_payment'] *
                                                               self.schedule.df['notional_discount_factor'])
        self.schedule.df['coupon_payment_pv'] = (self.schedule.df['coupon_payment'] *
                                                            self.schedule.df['coupon_discount_factor'])
        self.schedule.df['total_payment_pv'] = (self.schedule.df['notional_payment_pv'] +
                                                           self.schedule.df['coupon_payment_pv'])
        return self.schedule.df['total_payment_pv'].sum()

    def calc_clean_pv(self):
        return self.calc_dirty_pv() - self.calc_accrued_interest()

    def get_current_non_current_split(self, calc_npv=True):
        if calc_npv:
            self.calc_dirty_pv()

        current, non_current = 0, 0
        value_date_plus_one_year = frm.utils.VALUE_DATE + pd.DateOffset(years=1)

        if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
            current_period_idxs = self.schedule.df['notional_payment_date'] <= value_date_plus_one_year
        else:
            current_period_idxs = self.schedule.df['notional_payment_date'] < value_date_plus_one_year
        current += self.schedule.df['notional_payment_pv'][current_period_idxs].sum()
        non_current += self.schedule.df['notional_payment_pv'][~current_period_idxs].sum()

        if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
            current_period_idxs = self.schedule.df['coupon_payment_date'] <= value_date_plus_one_year
        else:
            current_period_idxs = self.schedule.df['coupon_payment_date'] < value_date_plus_one_year
        current += self.schedule.df['coupon_payment_pv'][current_period_idxs].sum()
        non_current += self.schedule.df['notional_payment_pv'][~current_period_idxs].sum()

        return current, non_current

    def _bucket_payments_helper(self, buckets, prefix):
        payment_col = prefix + 'payment'
        payment_date_col = prefix + 'payment_date'

        rows = []
        if frm.utils.INCLUDE_PAYMENTS_ON_VALUE_DATE_IN_NPV:
            mask_lower = self.schedule.df[payment_date_col] >= frm.utils.VALUE_DATE
        else:
            mask_lower = self.schedule.df[payment_date_col] > frm.utils.VALUE_DATE

        for i,v in enumerate(buckets):
            upper_pillar = frm.utils.VALUE_DATE + pd.DateOffset(years=v)
            mask_upper = self.schedule.df[payment_date_col] < (upper_pillar)
            rows.append(self.schedule.df[payment_col][mask_lower & mask_upper].sum())
            mask_lower = self.schedule.df[payment_date_col] >= upper_pillar

        mask_lower = self.schedule.df[payment_date_col] >= upper_pillar
        rows.append(self.schedule.df[payment_col][mask_lower].sum())
        index = buckets + (np.inf,)

        return pd.DataFrame(rows, index=index, columns=[payment_col])

    def bucket_payments(self, calc_npv=True, buckets=(1,2,3,5)):
        if calc_npv:
            self.calc_dirty_pv()
        coupon_payments_bucketed = self._bucket_payments_helper(buckets, prefix='coupon_')
        notional_payments_bucketed = self._bucket_payments_helper(buckets, prefix='notional_')
        combined = pd.concat([coupon_payments_bucketed, notional_payments_bucketed], axis=1)
        combined['total_payment'] = combined['coupon_payment'] + combined['notional_payment']
        return combined


    def calc_dv01(self):
        # Shared DV01 calculation with +/- 100 bps shift
        pass



@dataclass
class FixedLeg(Leg):
    fixed_rate: float | np.ndarray = np.nan

    def __post_init__(self):
        super().__post_init__()
        self._add_contractual_coupon_param(self.fixed_rate)
        #renamed_columns = {'coupon_contractual_component': self._get_solved_field_name()}
        #self.schedule.df = self.schedule.df.rename(columns=renamed_columns)
        #self.schedule.df[self._get_solved_field_name()] = self.fixed_rate

    def calc_coupon_payment(self, coupon_override=None):
        if coupon_override is not None:
            self.schedule.df['fixed_rate'] = coupon_override
        self.schedule.df['coupon_payment'] = self.pay_rcv.multiplier * self.schedule.df['notional'] * self.schedule.df['fixed_rate'] * self.schedule.df['period_yearfrac']
        mask = pd.isna(self.schedule.df['coupon_payment_date'])
        self.schedule.df.loc[mask,'coupon_payment'] = 0.0

    def _calc_accrued_interest(self):
        current_period_idxs = (self.schedule.df['period_start'] <= frm.utils.VALUE_DATE) & (self.schedule.df['period_end'] > frm.utils.VALUE_DATE)
        if current_period_idxs.sum() > 1:
            warnings.warn('Multiple periods found for accrual interest calculation.')
        accrued_period_yearfrac = year_frac(self.schedule.df['period_start'][current_period_idxs].iloc[0],
                                         frm.utils.VALUE_DATE,
                                         day_count_basis=self.day_count_basis)
        accrued_interest = (self.schedule.df['notional'][current_period_idxs]
                            * self.schedule.df['fixed_rate'][current_period_idxs]
                            * accrued_period_yearfrac).sum()
        return accrued_interest

    def _get_solved_field_name(self):
        return 'fixed_rate'

    def _get_legtype_specific_schedule_cols(self):
        return ['fixed_rate']


@dataclass
class FloatTermLeg(Leg):
    spread: float | np.ndarray = np.nan
    forward_rate_type: TermRate=TermRate.SIMPLE
    term_swap_curve: Optional[TermSwapCurve]=None

    def __post_init__(self):
        super().__post_init__()
        self._add_contractual_coupon_param(self.spread)
        # self.schedule.df[self._get_legtype_specific_schedule_cols()] = np.nan
        # renamed_columns = {'coupon_contractual_component': self._get_solved_field_name()}
        # self.schedule.df = self.schedule.df.rename(columns=renamed_columns)
        # self.schedule.df[self._get_solved_field_name()] = self.spread
        #self.calc_notional_schedule(coupon_fields_to_set_to_zero=['fixing','spread','coupon_rate'])

    def calc_coupon_payment(self, coupon_override=None):
        if self.term_swap_curve is None:
            raise MarketDataNotAvailableError('Term Swap curve not available.')

        self.schedule.df['fixing'] = np.nan
        if coupon_override is not None:
            self.schedule.df['spread'] = coupon_override
        self.schedule.df['coupon_rate'] = self.schedule.df['fixing'] + self.schedule.df['spread']
        self.schedule.df['coupon_payment'] = self.schedule.df['notional'] * self.schedule.df['coupon_rate'] * self.schedule.df['period_yearfrac']
        mask = pd.isna(self.schedule.df['coupon_payment_date'])
        self.schedule.df.loc[mask,'coupon_payment'] = 0.0

    def _calc_accrued_interest(self):
        current_period_idxs = (self.schedule.df['period_start'] <= frm.utils.VALUE_DATE) & (self.schedule.df['period_end'] > frm.utils.VALUE_DATE)
        if current_period_idxs.sum() > 1:
            warnings.warn('Multiple periods found for accrual interest calculation.')
        accrued_period_yearfrac = year_frac(self.schedule.df['period_start'][current_period_idxs].iloc[0], frm.utils.VALUE_DATE, day_count_basis=self.day_count_basis)
        accrued_interest = (self.schedule.df['notional'][current_period_idxs]
                            * self.schedule.df['spread'][current_period_idxs]
                            * accrued_period_yearfrac).sum()
        return accrued_interest

    def _get_solved_field_name(self):
        return 'spread'

    def _get_legtype_specific_schedule_cols(self):
        return ['fixing_date','fixing','spread','coupon_rate']


@dataclass
class FloatRFRLeg(Leg):
    spread: float | np.ndarray = np.nan
    forward_rate_type: RFRFixingCalcMethod=RFRFixingCalcMethod.DAILY_COMPOUNDED
    compound_spread: bool=False # TODO
    rfr_swap_curve: Optional[RFRSwapCurve]=None

    def __post_init__(self):
        super().__post_init__()
        self._add_contractual_coupon_param(self.spread)
        # self.schedule.df[self._get_legtype_specific_schedule_cols()] = np.nan
        # renamed_columns = {'coupon_contractual_component': self._get_solved_field_name()}
        # self.schedule.df = self.schedule.df.rename(columns=renamed_columns)
        # self.schedule.df[self._get_solved_field_name()] = self.spread
        #self.calc_notional_schedule(coupon_fields_to_set_to_zero=['fixing','spread','coupon_rate'])

    def calc_coupon_payment(self, coupon_override=None):
        if self.rfr_swap_curve is None:
            raise MarketDataNotAvailableError('OIS curve not available.')

        self.schedule.df['fixing'] = np.nan # TODO
        if coupon_override is not None:
            self.schedule.df['spread'] = coupon_override
        self.schedule.df['coupon_rate'] = self.schedule.df['fixing'] + self.schedule.df['spread']
        self.schedule.df['coupon_payment'] = self.schedule.df['notional'] * self.schedule.df['coupon_rate'] * self.schedule.df['period_yearfrac']
        mask = pd.isna(self.schedule.df['coupon_payment_date'])
        self.schedule.df.loc[mask,'coupon_payment'] = 0.0

    def _calc_accrued_interest(self):
        current_period_idxs = (self.schedule.df['period_start'] <= frm.utils.VALUE_DATE) & (self.schedule.df['period_end'] > frm.utils.VALUE_DATE)
        if current_period_idxs.sum() > 1:
            warnings.warn('Multiple periods found for accrual interest calculation.')
        accrued_period_yearfrac = year_frac(self.schedule.df['period_start'][current_period_idxs].iloc[0], frm.utils.VALUE_DATE, day_count_basis=self.day_count_basis)
        accrued_interest = (self.pay_rcv.multiplier * self.schedule.df['notional'][current_period_idxs]
                            * self.schedule.df['coupon_rate'][current_period_idxs]
                            * accrued_period_yearfrac).sum()
        return accrued_interest

    def _get_solved_field_name(self):
        return 'spread'

    def _get_legtype_specific_schedule_cols(self):
        return ['fixing','spread','coupon_rate']


@dataclass
class ZerocouponLeg(Leg):
    zero_rate: float=np.nan
    compounding_freq: CompoundingFreq=CompoundingFreq.ANNUAL

    def __post_init__(self):
        super().__post_init__()
        self._add_contractual_coupon_param(self.zero_rate)
        # renamed_columns = {'coupon_contractual_component': self._get_solved_field_name()}
        # self.schedule.df = self.schedule.df.rename(columns=renamed_columns)
        # self.schedule.df['zero_rate'] = self.zero_rate
        # TODO support specifying the terminal notional

    def calc_coupon_payment(self, coupon_override=None):
        if coupon_override is not None:
            self.schedule.df['zero_rate'] = coupon_override
        multiplier = (1.0 / discount_factor_from_zero_rate(
            years=self.schedule.df['period_yearfrac'],
            zero_rate=self.schedule.df['zero_rate'],
            compounding_freq=self.compounding_freq)) - 1.0
        self.schedule.df['coupon_payment'] = self.pay_rcv.multiplier * self.schedule.df['notional'] * multiplier
        mask = pd.isna(self.schedule.df['coupon_payment_date'])
        self.schedule.df.loc[mask,'coupon_payment'] = 0.0

    def _calc_accrued_interest(self):
        current_period_idxs = (self.schedule.df['period_start'] <= frm.utils.VALUE_DATE) & (
                    self.schedule.df['period_end'] > frm.utils.VALUE_DATE)
        if current_period_idxs.sum() > 1:
            warnings.warn('Multiple periods found for accrual interest calculation.')
        accrued_period_yearfrac = year_frac(self.schedule.df['period_start'][current_period_idxs].iloc[0], frm.utils.VALUE_DATE,
                                         day_count_basis=self.day_count_basis)
        multiplier = (1.0 / discount_factor_from_zero_rate(
            years=accrued_period_yearfrac,
            zero_rate=self.schedule.df['zero_rate'],
            compounding_freq=self.compounding_freq)) - 1.0
        accrued_interest =  self.pay_rcv.multiplier * self.schedule.df['notional'] * multiplier
        return accrued_interest

    def _get_solved_field_name(self):
        return 'zero_rate'

    def _get_legtype_specific_schedule_cols(self):
        return ['zero_rate']

# TBC at later date.
class InflationZCLeg(Leg):
    fixing_lag_months: int=2
    initial_fixing: float=np.nan

    def __post_init__(self):
        super().__post_init__()
        #self._add_contractual_coupon_param(self.spread)


    def calc_coupon_cashflow(self):
        pass

    def _calc_accrued_interest(self):
        pass

    def _get_solved_field_name(self):
        return 'spread' # Or final fixing?

# TBC at later date.
class InflationYoYLeg(Leg):
    fixing_lag_months: int=2
    initial_fixing: float=np.nan
    # forward_rate_type: TBC

    def __post_init__(self):
        super().__post_init__()
        #self._add_contractual_coupon_param(self.spread)

    def calc_coupon_cashflow(self):
        pass

    def _calc_accrued_interest(self):
        pass

    def _get_solved_field_name(self):
        return 'spread'


